/*
====================================================================================================

    Copyright (C) 2021 RRe36

    All Rights Reserved unless otherwise explicitly stated.


    By downloading this you have agreed to the license and terms of use.
    These can be found inside the included license-file
    or here: https://rre36.com/copyright-license

    Violating these terms may be penalized with actions according to the Digital Millennium
    Copyright Act (DMCA), the Information Society Directive and/or similar laws
    depending on your country.

====================================================================================================
*/

/* RENDERTARGETS: 3,5,14 */
layout(location = 0) out vec4 GBuffer;
layout(location = 1) out vec4 indirectLight;
layout(location = 2) out vec3 flatNormals;

#include "/lib/head.glsl"
#include "/lib/util/encoders.glsl"

const bool shadowHardwareFiltering = true;

in vec2 uv;

#ifndef DIM
flat in mat4x3 lightColor;
#define blocklightColor lightColor[3]
#elif DIM == -1
flat in mat2x3 lightColor;
#define blocklightColor lightColor[1]
#elif DIM == 1
flat in mat3x3 lightColor;
#define blocklightColor lightColor[2]
#endif

#if DIM == -1
    #define NO_DIRECTLIGHT
#endif

uniform sampler2D colortex0;
uniform sampler2D colortex1;
uniform sampler2D colortex2;
uniform sampler2D colortex3;
uniform sampler2D colortex4;
uniform sampler2D colortex5;
uniform sampler2D colortex7;

uniform sampler2D depthtex0;
uniform sampler2D depthtex2;

uniform sampler2D noisetex;

uniform sampler2DShadow shadowtex1;
uniform sampler2D shadowcolor0;
uniform sampler2D shadowcolor1;

uniform int frameCounter;

uniform float aspectRatio;
uniform float near, far;
uniform float lightFlip;
uniform float sunAngle;

uniform vec2 skyCaptureResolution;
uniform vec2 pixelSize, viewSize;
uniform vec2 taaOffset;

uniform vec3 upDir, cameraPosition;
uniform vec3 sunDir, moonDir, lightDir;

uniform mat4 gbufferModelView, gbufferModelViewInverse;
uniform mat4 gbufferProjection, gbufferProjectionInverse;
uniform mat4 shadowModelView, shadowModelViewInverse;
uniform mat4 shadowProjection, shadowProjectionInverse;

/* ------ includes ------*/
#define FUTIL_MAT16
#define FUTIL_LINDEPTH
#define FUTIL_LIGHTMAP
#include "/lib/fUtil.glsl"
#include "/lib/util/transforms.glsl"
#include "/lib/frag/bluenoise.glsl"
#include "/lib/frag/gradnoise.glsl"

#include "/lib/atmos/project.glsl"

#include "/lib/brdf/fresnel.glsl"
#include "/lib/brdf/hammon.glsl"


/* ------ DITHER ------ */

#define HASHSCALE1 .1031
float hash12(vec2 x) {
    vec3 x3  = fract(vec3(x.xyx) * HASHSCALE1);
        x3  += dot(x3, x3.yzx + 19.19);
    return fract((x3.x + x3.y) * x3.z);
}

#define HASHSCALE3 vec3(.1031, .1030, .0973)
vec2 hash22(vec2 p) {
	vec3 p3 = fract(vec3(p.xyx) * HASHSCALE3);
    p3 += dot(p3, p3.yzx + 19.19);
    return fract((p3.xx + p3.yz) * p3.zy);
}

float ditherHashNoise() {
    float noise     = hash12(gl_FragCoord.xy);
        noise   = fract(noise + (frameCounter * 7.0 * rcp(255.0)) *0);

    return noise;
}
float ditherBluenoise2() {
    ivec2 coord = ivec2(fract(gl_FragCoord.xy/256.0)*256.0);
    float noise = texelFetch(noisetex, coord, 0).a;

        noise   = fract(noise + (frameCounter * 7.0 * rcp(255.0)));

    return noise;
}


/* ------ SAMPLE DIST ------ */
/*
//Fermats Spiral and Spheremap functions from raspberry by rutherin https://www.patreon.com/user?u=12731730
vec2 fermatsSpiralGoldenN(float index, float total) {
	float theta = index * goldenAngle;
	return vec2(sin(theta), cos(theta)) * sqrt(index / total);
}
vec2 fermatsSpiralGoldenS(float index, float total) {
	float theta = index * goldenAngle;
	return vec2(sin(theta), cos(theta)) * sqr(index / total);
}

vec3 sphereMap(vec2 a) {
    float phi = a.y * 2.0 * pi;
    float cosTheta = 1.0 - a.x;
    float sinTheta = sqrt(1.0 - cosTheta * cosTheta);

    return vec3(cos(phi) * sinTheta, sin(phi) * sinTheta, cosTheta);
}*/


vec3 genUnitVector(vec2 p) {
    p.x *= tau; p.y = p.y * 2.0 - 1.0;
    return vec3(sincos(p.x) * sqrt(1.0 - p.y * p.y), p.y);
}
vec3 GenerateCosineVectorSafe(vec3 vector, vec2 xy) {
	// Apparently this is actually this simple.
	// http://www.amietia.com/lambertnotangent.html
	// (cosine lobe around vector = lambertian BRDF)
	// This one deals with ther rare case where cosineVector == (0,0,0)
	// Can just normalize it instead if you are sure that won't be a problem
	vec3 cosineVector = vector + genUnitVector(xy);
	float lenSq = dot(cosineVector, cosineVector);
	return lenSq > 0.0 ? cosineVector * inversesqrt(lenSq) : vector;
}

/* ------ SSPT ------ */

#ifdef ssptEnabled

vec3 screenspaceRT(vec3 position, vec3 direction, float noise) {
    const uint maxSteps     = 8;
    const float stepSize    = halfPi / float(maxSteps);

    vec3 stepVector         = direction * stepSize;

    vec3 endPosition        = position + stepVector * maxSteps;
    vec3 endScreenPosition  = viewToScreenSpace(endPosition);

    vec2 maxPosXY           = max(abs(endScreenPosition.xy * 2.0 - 1.0), vec2(1.0));
    float stepMult          = minOf(vec2(1.0) / maxPosXY);
        stepVector         *= stepMult;

    // closest texel iteration
    vec3 samplePos          = position;
    
        samplePos          += stepVector / 6.0;
    vec3 screenPos          = viewToScreenSpace(samplePos);

    const float thresh      = 0.25;

    if (saturate(screenPos.xy) == screenPos.xy) {
        float depthSample   = texelFetch(depthtex0, ivec2(screenPos.xy * viewSize), 0).x;
        float linearSample  = depthLinear(depthSample);
        float currentDepth  = depthLinear(screenPos.z);

        if (linearSample < currentDepth) {
            float dist      = abs(linearSample - currentDepth) / clamp(currentDepth, 0.25 / far, 2.0 / far);
            if (dist <= thresh) return vec3(screenPos.xy, depthSample);
        }
    }
    
        samplePos          += stepVector * noise;

    for (uint i = 0; i < maxSteps; ++i) {
        vec3 screenPos      = viewToScreenSpace(samplePos);
            samplePos      += stepVector;
        if (saturate(screenPos.xy) != screenPos.xy) break;

        float depthSample   = texelFetch(depthtex0, ivec2(screenPos.xy * viewSize), 0).x;
        float linearSample  = depthLinear(depthSample);
        float currentDepth  = depthLinear(screenPos.z);

        if (linearSample < currentDepth) {
            float dist      = abs(linearSample - currentDepth) / clamp(currentDepth, 0.25 / far, 2.0 / far);
            if (dist <= thresh) return vec3(screenPos.xy, depthSample);
        }
    }

    return vec3(1.1);
}

#include "/lib/voxel/store.glsl"

#include "/lib/voxel/lpvSampling.glsl"

bool IntersectAABox(vec3 rayPos, vec3 invRayDir, vec3 minBounds, vec3 maxBounds, inout float dist) {
	// This will not intersect backfaces, or in other words this will not intersect the box from inside it.
	// You should use the version that rakes in radii instead of min & max bounds if you can, as it will be faster.
	vec3 minBoundsDist = (minBounds - rayPos) * invRayDir;
	vec3 maxBoundsDist = (maxBounds - rayPos) * invRayDir;

	vec3 minDists = min(minBoundsDist, maxBoundsDist);
	vec3 maxDists = max(minBoundsDist, maxBoundsDist);

	float maxOfMin = maxOf(minDists); // distance to frontfaces
	float minOfMax = minOf(maxDists); // distance to backfaces

	dist = maxOfMin; // Distance to intersection (if the box was intersected)

	// Check for intersection & return whether or not we intersected
	return dist > 0.0 && maxOfMin <= minOfMax;
}
bool IntersectAABox(vec3 rayPos, vec3 invRayDir, vec3 minBounds, vec3 maxBounds, out float dist, inout vec3 hitNormal) {
	// This will not intersect backfaces, or in other words this will not intersect the box from inside it.
	// You should use the version that rakes in radii instead of min & max bounds if you can, as it will be faster.
	vec3 minBoundsDist = (minBounds - rayPos) * invRayDir;
	vec3 maxBoundsDist = (maxBounds - rayPos) * invRayDir;

	vec3 minDists = min(minBoundsDist, maxBoundsDist);
	vec3 maxDists = max(minBoundsDist, maxBoundsDist);

	float maxOfMin = maxOf(minDists); // distance to frontfaces
	float minOfMax = minOf(maxDists); // distance to backfaces

	dist = maxOfMin; // Distance to intersection (if the box was intersected)

	// Determine normal
	if (dist == minDists.x) {
		hitNormal = vec3(invRayDir.x < 0.0 ? 1.0 : -1.0, 0.0, 0.0);
	} else if (dist == minDists.y) {
		hitNormal = vec3(0.0, invRayDir.y < 0.0 ? 1.0 : -1.0, 0.0);
	} else {
		hitNormal = vec3(0.0, 0.0, invRayDir.z < 0.0 ? 1.0 : -1.0);
	}

	// Check for intersection & return whether or not we intersected
	return dist > 0.0 && maxOfMin <= minOfMax;
}

bool IntersectAABox(vec3 rayPos, vec3 invRayDir, vec3 boxRadii, out float dist, inout vec3 hitNormal) {
	// This one will not intersect backfaces, or in other words this will not intersect the box from inside it.
	vec3 scaledRayPos = rayPos * invRayDir;
	vec3 minDists = -boxRadii * abs(invRayDir) - scaledRayPos; // distance to each frontfacing plane
	vec3 maxDists =  boxRadii * abs(invRayDir) - scaledRayPos; // distance to each backfacing plane

	float maxOfMin = maxOf(minDists); // distance to the frontfaces
	float minOfMax = minOf(maxDists); // distance to the backfaces

	dist = maxOfMin; // Distance to intersection (if the box was intersected)

	// Determine normal
	if (dist == minDists.x) {
		hitNormal = vec3(invRayDir.x < 0.0 ? 1.0 : -1.0, 0.0, 0.0);
	} else if (dist == minDists.y) {
		hitNormal = vec3(0.0, invRayDir.y < 0.0 ? 1.0 : -1.0, 0.0);
	} else {
		hitNormal = vec3(0.0, 0.0, invRayDir.z < 0.0 ? 1.0 : -1.0);
	}

	// Check for intersection & return whether or not we intersected
	return dist > 0.0 && maxOfMin <= minOfMax;
}

bool intersectVoxel(vec3 relOrigin, vec3 dir, inout float hitDist) {
    vec3 invDir     = 1.0 / dir;

    vec3 minBounds  = vec3(0);
    vec3 maxBounds  = vec3(1);
    bool intersected = IntersectAABox(relOrigin, invDir, minBounds, maxBounds, hitDist);

    return intersected;
}
bool intersectVoxel(vec3 relOrigin, vec3 dir, inout float hitDist, inout vec3 hitNormal) {
    vec3 invDir     = 1.0 / dir;

    vec3 minBounds  = vec3(0);
    vec3 maxBounds  = vec3(1);
    bool intersected = IntersectAABox(relOrigin, invDir, minBounds, maxBounds, hitDist, hitNormal);

    return intersected;
}

bool IntersectSphere(vec3 relOrigin, vec3 dir, float radius, out float hitDist, inout vec3 hitNormal) {
	vec3 s0 = vec3(0.5, 0.5, 0.5); // Sphere center relative to voxel

    float a = dot(dir, dir);
    vec3 s0_r0 = relOrigin - s0;
    float b = 2.0 * dot(dir, s0_r0);
    float c = dot(s0_r0, s0_r0) - (radius * radius);

    if (b*b - 4.0*a*c < 0.0) {
        return false;
    }

    hitDist = (-b - sqrt((b*b) - 4.0*a*c))/(2.0*a);

    hitNormal = -dir;

	return true;
}

bool intersectVoxel(vec3 relOrigin, vec3 dir, int ID, inout float hitDist, inout vec3 hitNormal) {
    vec3 invDir     = 1.0 / dir;

    vec3 minBounds  = vec3(0);
    vec3 maxBounds  = vec3(1);

    if (ID == 2) {
        return IntersectSphere(relOrigin, dir, 0.4, hitDist, hitNormal);
    }

    return IntersectAABox(relOrigin, invDir, minBounds, maxBounds, hitDist, hitNormal);
}


bool marchVoxel(vec3 pos, ivec3 index, vec3 dir, const int maxDistance, out vec3 hitPos, out vec3 hitNormal, out ivec3 hitIndex) {
    //pos -= vec3(1,0,1);     // no idea why its offset like this
    //index -= ivec3(1,0,1);    

    if (outsideVoxelVolume(index)) return false;

    ivec3 startIndex = index;

	vec3 nextT;
	vec3 deltaDist = 1.0 / abs(dir);
	ivec3 deltaSign;

    hitIndex = ivec3(-1);

	for (int axis = 0; axis < 3; ++axis) {
		if (dir[axis] < 0.0) {
			deltaSign[axis] = -1;
			nextT[axis] = (pos[axis] - index[axis]) * deltaDist[axis];
		} else if (dir[axis] > 0.0) {
			deltaSign[axis] = 1;
			nextT[axis] = (index[axis] - pos[axis]) * deltaDist[axis] + deltaDist[axis];
		} else {
			deltaSign[axis] = 0;
			nextT[axis] = uintBitsToFloat(0x7f800000u);// = infinity
		}
	}

	if (deltaSign.x == 0 && deltaSign.y == 0 && deltaSign.z == 0) { return false; }

	bool hit = false;
    int i       = 0;
	do {
        float hitDist;

		float minComp = minOf(nextT);
		if (nextT.x == minComp) {
			hitDist   = nextT.x;
			hitNormal = vec3(-deltaSign.x, 0.0, 0.0);

			index.x   += deltaSign.x;
			nextT.x += deltaDist.x;
		}
		if (nextT.y == minComp) {
			hitDist   = nextT.y;
			hitNormal = vec3(0.0, -deltaSign.y, 0.0);

			index.y   += deltaSign.y;
			nextT.y += deltaDist.y;
		}
		if (nextT.z == minComp) {
			hitDist   = nextT.z;
			hitNormal = vec3(0.0, 0.0, -deltaSign.z);

			index.z   += deltaSign.z;
			nextT.z += deltaDist.z;
		}

		if (outsideVoxelVolume(index)) break;

		hit = getVoxelOccupancy(index);

        if (hit) {
            hit     = intersectVoxel(pos - index, dir, hitDist, hitNormal);
            hitPos  = pos + hitDist * dir;
            hitIndex = index;
        }

	} while (++i < maxDistance && !hit);

    return hit;
}

bool marchVoxel(vec3 pos, ivec3 index, vec3 endPos, ivec3 endIndex, inout vec3 hitPos, inout vec3 hitNormal, out ivec3 hitIndex) {
    //pos -= vec3(1,0,1);     // no idea why its offset like this
    //index -= ivec3(1,0,1);
    //endPos -= vec3(1,0,1);
    //endIndex -= ivec3(1,0,1);
    

    if (outsideVoxelVolume(index)) return false;
    //if (distance(index, endIndex) < 0.1) return false;

    vec3 dir    = normalize(endPos - pos);

    ivec3 startIndex = index;
    float endDistance = distance(index, endIndex) + 1.0;

	vec3 nextT;
	vec3 deltaDist = 1.0 / abs(dir);
	ivec3 deltaSign;

    hitIndex = ivec3(-1);

	for (int axis = 0; axis < 3; ++axis) {
		if (dir[axis] < 0.0) {
			deltaSign[axis] = -1;
			nextT[axis] = (pos[axis] - index[axis]) * deltaDist[axis];
		} else if (dir[axis] > 0.0) {
			deltaSign[axis] = 1;
			nextT[axis] = (index[axis] - pos[axis]) * deltaDist[axis] + deltaDist[axis];
		} else {
			deltaSign[axis] = 0;
			nextT[axis] = uintBitsToFloat(0x7f800000u);// = infinity
		}
	}

	if (deltaSign.x == 0 && deltaSign.y == 0 && deltaSign.z == 0) { return false; }

	bool hit = false;
    int i       = 0;
	do {
        float hitDist;

		float minComp = minOf(nextT);
		if (nextT.x == minComp) {
			hitDist   = nextT.x;
			hitNormal = vec3(-deltaSign.x, 0.0, 0.0);

			index.x   += deltaSign.x;
			nextT.x += deltaDist.x;
		}
		if (nextT.y == minComp) {
			hitDist   = nextT.y;
			hitNormal = vec3(0.0, -deltaSign.y, 0.0);

			index.y   += deltaSign.y;
			nextT.y += deltaDist.y;
		}
		if (nextT.z == minComp) {
			hitDist   = nextT.z;
			hitNormal = vec3(0.0, 0.0, -deltaSign.z);

			index.z   += deltaSign.z;
			nextT.z += deltaDist.z;
		}

		if (outsideVoxelVolume(index)) break;

        if (distance(startIndex, index) > endDistance) break;

        //if (index == startIndex) continue;

		hit = getVoxelOccupancy(index);

        if (hit) {
            hit     = intersectVoxel(pos - index, dir, hitDist, hitNormal);
            hitPos  = pos + hitDist * dir;
            hitIndex = index;
        }

	} while (++i < 64 && !hit);

    return hit;
}

vec3 getVoxelTint(ivec3 index) {
    ivec2 uv    = getVoxelPixel(index);

    vec4 voxel  = texelFetch(shadowcolor0, uv, 0);
    vec4 albedo = vec4(unpack2x8(voxel.x), unpack2x8(voxel.y));

    return unpack4x4(voxel.a).x > 0.5 ? albedo.rgb : vec3(1.0);
}


vec3 decodeVoxelTint(vec4 voxel) {
    vec4 albedo = vec4(unpack2x8(voxel.x), unpack2x8(voxel.y));

    return unpack4x4(voxel.a).x > 0.5 ? albedo.rgb : vec3(1.0);
}
/*
bool marchVoxel(vec3 pos, ivec3 index, vec3 endPos, ivec3 endIndex, out vec3 voxelTint) {
    //pos -= vec3(1,0,1);     // no idea why its offset like this
    //index -= ivec3(1,0,1);
    //endPos -= vec3(1,0,1);
    //endIndex -= ivec3(1,0,1);

    voxelTint    = vec3(0);
    

    if (outsideVoxelVolume(index)) return false;
    if (distance(index, endIndex) < 0.1) return false;

    vec3 dir    = normalize(endPos - pos);

    ivec3 startIndex = index;
    float endDistance = distance(index, endIndex) + 0.5;

	vec3 nextT;
	vec3 deltaDist = 1.0 / abs(dir);
	ivec3 deltaSign;

	for (int axis = 0; axis < 3; ++axis) {
		if (dir[axis] < 0.0) {
			deltaSign[axis] = -1;
			nextT[axis] = (pos[axis] - index[axis]) * deltaDist[axis];
		} else if (dir[axis] > 0.0) {
			deltaSign[axis] = 1;
			nextT[axis] = (index[axis] - pos[axis]) * deltaDist[axis] + deltaDist[axis];
		} else {
			deltaSign[axis] = 0;
			nextT[axis] = uintBitsToFloat(0x7f800000u);// = infinity
		}
	}

	if (deltaSign.x == 0 && deltaSign.y == 0 && deltaSign.z == 0) { return false; }

	bool hit = false;
    int i       = 0;
	do {
        float hitDist;

		float minComp = minOf(nextT);
		if (nextT.x == minComp) {
			hitDist   = nextT.x;

			index.x   += deltaSign.x;
			nextT.x += deltaDist.x;
		}
		if (nextT.y == minComp) {
			hitDist   = nextT.y;

			index.y   += deltaSign.y;
			nextT.y += deltaDist.y;
		}
		if (nextT.z == minComp) {
			hitDist   = nextT.z;

			index.z   += deltaSign.z;
			nextT.z += deltaDist.z;
		}

		if (outsideVoxelVolume(index)) break;

        //if (distance(startIndex, index) > endDistance) break;

        //if (index == endIndex || index == startIndex) continue;

		hit = getVoxelOccupancy(index);

	} while (++i < 256 && !hit);

    if (hit) {
        voxelTint = getVoxelTint(index);
    }

    return hit;
}

vec3 debugVoxel(vec3 viewPos, vec3 flatSceneNormal) {
    vec3 scenePos   = sceneToVoxelSpace(viewToSceneSpace(viewPos));

    ivec3 endVoxel  = voxelSpaceToIndex(scenePos - flatSceneNormal / 128.0);
    
    vec3 startPos   = sceneToVoxelSpace(vec3(0));

    ivec3 startIndex = voxelSpaceToIndex(startPos);

    vec3 tint;

    bool hit    = marchVoxel(startPos, startIndex, scenePos, endVoxel, tint);

    return tint;
}*/

#include "/lib/light/warp.glsl"

vec3 getShadowmapPos(vec3 scenePos, const float bias) {  //shadow 2d
    vec3 pos    = scenePos + vec3(bias) * lightDir;
    float a     = length(pos);
        pos     = transMAD(shadowModelView, pos);
        pos     = projMAD(shadowProjection, pos);
        pos.z  *= 0.2;
        pos.z  -= 0.0012*(saturate(a/256.0));

        pos.xy  = shadowmapWarp(pos.xy);

        pos.x   = pos.x * 0.5 + 0.5;
        pos.xy  = pos.yx;

    return pos*0.5+0.5;
}

vec3 getDirectLight(vec3 scenePos, vec3 sceneNormal) {
    float sDotN     = dot(lightDir, sceneNormal);
    if (sDotN <= 0.0) return vec3(0.0);

    float shadow    = 1.0;

    vec3 shadowPos  = getShadowmapPos(scenePos, 0.08);
        shadow  *= texture(shadowtex1, shadowPos);

    return texture(shadowcolor0, shadowPos.xy).rgb * (shadow * saturate(sDotN));
}

ivec3 scaleIndex(ivec3 index, int targetLod) {
    vec3 x  = vec3(index) / float(targetLod);

    return ivec3(floor(x));
}
bool marchBVH(vec3 pos, ivec3 index, vec3 dir, out vec3 hitPos, out vec3 hitNormal, out ivec3 hitIndex, inout vec4 voxel) {
    if (outsideVoxelVolume(index)) return false;

    ivec3 deltaSign     = ivec3(sign(dir));

    if(all(equal(deltaSign, ivec3(0)))) return false;

    ivec3 positiveDir   = max(deltaSign, ivec3(0));
    ivec3 negativeDir   = min(deltaSign, ivec3(0));

    vec3 deltaT         = 1.0 / abs(dir);

    const int maxLevel  = OCTREE_LEVELS - 1;

    vec3 steps          = (index - pos) * deltaSign + positiveDir - 1;

    int level           = 0;

    while(level++ < maxLevel && !getVoxelOccupancy(index, level));

    --level;

    bool hit            = false;
    int iterations      = 0;

    do {
        float minT;
        bvec3 isMinComp;

        bool potentialHit   = false;
        int octreeIterations = 0;

        if (outsideVoxelVolume(index)) break;

        do {
            int size        = int(pow(2, level));
            ivec3 plane     = (scaleIndex(index, size) + positiveDir) * size;
            vec3 planeT     = abs(plane - pos) * deltaT;

            minT            = minOf(planeT);

            if (minT > 256.0) break;

            isMinComp       = bvec3(planeT.x == minT, planeT.y == minT, planeT.z == minT);

            ivec3 prevIndex = index;

            ivec3 stepsToNext = ivec3(minT * abs(dir) - steps + 0.5 * vec3(isMinComp));
                steps      += stepsToNext;
                index      += stepsToNext * deltaSign;

            if (outsideVoxelVolume(index)) break;

            int levelAbove  = level + 1;
            int sizeAbove   = int(pow(2, levelAbove));

            bool stepUp     = scaleIndex(prevIndex, sizeAbove) != scaleIndex(index, sizeAbove);
                level       = min(stepUp ? levelAbove : level, maxLevel);

            bool blocked    = getVoxelOccupancy(index, level, voxel);

            while (blocked && level > 0) {
                blocked     = getVoxelOccupancy(index, --level, voxel);
            }

            potentialHit    = blocked && level == 0;
        } while(++octreeIterations < 64 && !potentialHit);

        if (potentialHit) {
            hitNormal       = vec3(isMinComp) * -deltaSign;

            float hitDist   = minT;

            ivec2 voxelID   = unpack2x8I(voxel.z);

            //if (voxelID.y == 2) continue;

            hit             = intersectVoxel(pos - index, dir, voxelID.y, hitDist, hitNormal);
            hitPos          = pos + hitDist * dir;

            hitIndex        = index;
        }

    } while (++iterations < 64 && !hit);

    return hit;
}

vec3 getVoxelMultibounce(vec3 voxelPos, ivec3 voxelIndex, vec3 hitNormal, vec3 dir, inout bool hit) {
    //hit     = marchVoxel(voxelPos, voxelIndex, dir, 64, voxelPos, hitNormal, voxelIndex);
    vec4 voxel = vec4(0);
    hit     = marchBVH(voxelPos, voxelIndex, dir, voxelPos, hitNormal, voxelIndex, voxel);

    if (hit) {
        if (dot(dir, hitNormal) > 0.0) return vec3(0);

        ivec2 voxelID = unpack2x8I(voxel.z);

        vec3 voxelTint      = decodeVoxelTint(voxel);

        //return voxelTint * float(voxelID.x > 0);

        //return getLight(voxelToSceneSpace(voxelPos + vec3(1,0,1)) + hitNormal * 0.25) * voxelTint;
        return readLPV(voxelSpaceToIndex(voxelPos + hitNormal * 0.05)) * voxelTint;
    } else {
        return vec3(0);
    }
}

vec3 PTLight(vec3 viewPos, vec3 scenePos, vec3 sceneNormal, vec3 flatNormal, vec2 dither) {
    vec3 totalColor     = vec3(0.0);

    #ifndef NO_DIRECTLIGHT
        #ifndef DIM
        vec3 directCol      = (sunAngle<0.5 ? lightColor[0] : lightColor[1]) * lightFlip;
        #elif DIM == 1
        vec3 directCol      = lightColor[0];
        #endif
    #endif

    const float a1 = 1.0 / rho;
    const float a2 = a1 * a1;

    vec2 quasirandomCurr = 0.5 + fract(vec2(a1, a2) * frameCounter + 0.5);

    vec2 noiseCurr      = hash22(gl_FragCoord.xy + frameCounter);

    vec3 voxelPos       = sceneToVoxelSpace(scenePos);
    ivec3 startVoxel    = voxelSpaceToIndex(voxelPos - flatNormal / 128.0);
        voxelPos       += flatNormal / 128.0;

    vec3 direction      = normalize(scenePos);
    vec3 oldDirection   = direction;

    for (uint i = 0; i < ssptSPP; ++i) {
        ++quasirandomCurr;
        noiseCurr      += hash22(vec2(gl_FragCoord.xy + vec2(cos(quasirandomCurr.x), sin(quasirandomCurr.y))));

        vec2 vectorXY   = fract(sqrt(2.0) * quasirandomCurr + noiseCurr);

            direction   = GenerateCosineVectorSafe(sceneNormal, vectorXY);

        if (dot(sceneNormal, direction) < 0.0) direction = -direction;

        float brdf      = clamp(diffuseHammon(sceneNormal, -oldDirection, direction, 1.0) / saturate(dot(direction, sceneNormal)/pi), 0.0, halfPi);

        bool hit        = false;

        vec3 screenHit  = screenspaceRT(viewPos, normalize(mat3(gbufferModelView) * direction), dither.y);

            hit         = screenHit.z < 1.0;

        if (hit) {
            ivec2 hitPixel      = ivec2(screenHit.xy * viewSize);
            vec3 flatHitNormal  = texelFetch(colortex3, hitPixel, 0).xyz * 2.0 - 1.0;
            vec4 albedoSample   = texelFetch(colortex0, hitPixel, 0);

            vec3 hitPosView     = screenToViewSpace(screenHit);
            vec3 hitPosScene    = viewToSceneSpace(hitPosView);

            vec4 dataSample     = texelFetch(colortex5, hitPixel, 0);

            float emissionFalloff = 1.0 - linStep(distance(hitPosView, viewPos), ssptEmissionDistance / pi, ssptEmissionDistance);
                emissionFalloff = cube(emissionFalloff);

            if (dataSample.a > 1e-8) {
                totalColor     += dataSample.rgb * emissionFalloff * brdf * halfPi;
            }

            vec3 directLight    = texelFetch(colortex7, hitPixel, 0).rgb * 2.0;
            totalColor         += albedoSample.rgb * directLight * directCol * brdf * pi;

            totalColor         += readLPV(sceneToVoxelIndex(hitPosScene + flatHitNormal * 0.1)) * brdf * albedoSample.rgb * pi;
        } else {
            vec4 voxel  = vec4(0);
            vec3 hitNormal = vec3(0);

            hit         = marchBVH(voxelPos, startVoxel, direction, voxelPos, hitNormal, startVoxel, voxel);

            ivec2 voxelID = unpack2x8I(voxel.z);

            vec3 voxelTint = decodeVoxelTint(voxel);

            totalColor += readLPV(voxelSpaceToIndex(voxelPos + hitNormal * 0.05)) * voxelTint * brdf * pi;
            totalColor += getDirectLight(voxelToSceneSpace(voxelPos), hitNormal) * directCol * voxelTint * brdf * pi;
        }

        if (!hit) {
            #ifndef DIM
                float groundOcclusion = exp(-max0(-direction.y) * sqrPi);

                totalColor           += texture(colortex4, projectSky(direction, 1)).rgb * brdf / sqrt2;
                totalColor           += minimumAmbientColor * minimumAmbientMult * minimumAmbientIllum * brdf * (groundOcclusion * 0.5 + 0.5);

                #else

                #if DIM == -1
                totalColor           += lightColor[0] * brdf;
                #else
                totalColor           += lightColor[1] * brdf;
                #endif
            #endif
        }
    }

    return totalColor / ssptSPP;
}

mat3 raytraceIndirectLight(vec3 viewPos, vec3 sceneNormal, vec3 flatStartNormal, vec2 dither, float skyOcclusion, bool hand) {
    vec3 viewNormal     = mat3(gbufferModelView) * sceneNormal;

    vec3 skylight       = vec3(0.0);
    vec3 bouncedLight   = vec3(0.0);
    vec3 emission       = vec3(0.0);

    //return mat3(0);

    const float a1 = 1.0 / rho;
    const float a2 = a1 * a1;

    #ifndef NO_DIRECTLIGHT
    vec3 fakeBounceColor = (sunAngle<0.5 ? lightColor[0] : lightColor[1]) * lightFlip * 0.004;
        fakeBounceColor += lightColor[2] * 0.33;
    #endif

    //skyOcclusion        = 1.0;

    //#if ssptBounces <= 1

    /*
        Single Bounce 
    */

    vec2 quasirandomCurr = 0.5 + fract(vec2(a1, a2) * frameCounter + 0.5);

    vec2 noiseCurr      = hash22(gl_FragCoord.xy + frameCounter);

    vec3 voxelPos       = sceneToVoxelSpace(viewToSceneSpace(viewPos));
    ivec3 startVoxel    = voxelSpaceToIndex(voxelPos - flatStartNormal / 128.0);
        voxelPos       += flatStartNormal / 128.0;
    //ivec3 startIndex    = voxelSpaceToIndex(voxelPos - flatStartNormal / 64.0);

    for (uint i = 0; i < ssptSPP; ++i) {
        ++quasirandomCurr;
        noiseCurr      += hash22(vec2(gl_FragCoord.xy + vec2(cos(quasirandomCurr.x), sin(quasirandomCurr.y))));

        vec2 vectorXY   = fract(sqrt(2.0) * quasirandomCurr + noiseCurr);

        vec3 rayDirection   = GenerateCosineVectorSafe(sceneNormal, vectorXY);
        vec3 rayWorldDir    = rayDirection;
            rayDirection    = normalize(mat3(gbufferModelView) * rayDirection);

        if (dot(viewNormal, rayDirection) < 0.0) rayDirection = -rayDirection;

        vec3 hitPosition    = screenspaceRT(viewPos, rayDirection, dither.y);

        bool hit            = hitPosition.z < 1.0;

        float brdf          = clamp(diffuseHammon(viewNormal, -normalize(viewPos), rayDirection, 1.0) / saturate(dot(rayDirection, viewNormal)/pi), 0.0, halfPi);
            //brdf            = saturate(dot(rayDirection, viewNormal)) * sqrt2;
/*
        if (hit && false) {
            ivec2 hitPixel      = ivec2(hitPosition.xy * viewSize);
            vec3 flatHitNormal  = texelFetch(colortex3, hitPixel, 0).xyz * 2.0 - 1.0;
            vec4 albedoSample   = texelFetch(colortex0, hitPixel, 0);

            vec3 hitPosView     = screenToViewSpace(hitPosition);
            vec3 hitPosScene    = viewToSceneSpace(hitPosView);

            vec3 hitVoxel       = sceneToVoxelSpace(hitPosScene);

            vec3 voxelNormal    = flatHitNormal;
            ivec3 voxelIndex    = ivec3(0);

            bool voxelHit       = marchVoxel(voxelPos, startVoxel, hitVoxel, voxelSpaceToIndex(hitVoxel - flatHitNormal / 128.0), hitVoxel, voxelNormal, voxelIndex);

            if (voxelHit) {
                vec3 voxelTint      = getVoxelTint(voxelIndex);

                emission           += readLPV(voxelSpaceToIndex(hitVoxel + voxelNormal * 0.25)) * brdf * voxelTint * pi;
                //emission           += getLight(voxelToSceneSpace(hitVoxel + vec3(1,0,1)) + voxelNormal * 0.25) * voxelTint * brdf * pi;
                //bouncedLight       += getDirectLight(voxelToSceneSpace(hitVoxel + ivec3(1,0,1)), voxelNormal) * voxelTint * brdf;
            } else {
                vec4 dataSample     = texelFetch(colortex5, hitPixel, 0);

                float emissionFalloff = 1.0 - linStep(distance(hitPosView, viewPos), ssptEmissionDistance / pi, ssptEmissionDistance);
                    emissionFalloff = cube(emissionFalloff);

                if (dataSample.a > 1e-8) {
                    emission       += dataSample.rgb * emissionFalloff * brdf * halfPi;
                }

                emission           += readLPV(sceneToVoxelIndex(hitPosScene + flatHitNormal * 0.25)) * brdf * albedoSample.rgb * pi;
            }

            #ifndef NO_DIRECTLIGHT

            vec3 directLight    = texelFetch(colortex7, hitPixel, 0).rgb * 2.0;
            bouncedLight       += albedoSample.rgb * directLight * brdf;

            #endif
            
        } else {
            
        }*/

            //#ifdef VOXEL_MULTIBOUNCE
            emission   += getVoxelMultibounce(voxelPos, startVoxel, flatStartNormal, rayWorldDir, hit) * brdf * pi;
            //#endif

        if (!hit) {
            #ifndef DIM
                float groundOcclusion = exp(-max0(-rayWorldDir.y) * sqrPi);

                skylight           += mix(fakeBounceColor * pow4(skyOcclusion), texture(colortex4, projectSky(rayWorldDir, 1)).rgb * cube(skyOcclusion), groundOcclusion) * brdf;
                skylight           += minimumAmbientColor * minimumAmbientMult * minimumAmbientIllum * brdf * (groundOcclusion * 0.5 + 0.5);

                #else

                #if DIM == -1
                skylight           += lightColor[0] * brdf;
                #else
                skylight           += lightColor[1] * brdf;
                #endif
            #endif
        }
    }

    skylight       /= float(ssptSPP);
    bouncedLight   /= float(ssptSPP);
    emission       /= float(ssptSPP);

    emission       *= float(!hand);

    return mat3(skylight / sqrt3, bouncedLight * pi, emission);
}

#endif


/* ------ FALLBACK ------ */

/*
    SSAO based on BSL Shaders by Capt Tatsu with permission
*/

vec2 offsetDist(float x) {
	float n = fract(x * 8.0) * pi;
    return vec2(cos(n), sin(n)) * x;
}

float getDSSAO(sampler2D depthtex, vec3 sceneNormal, float depth, vec2 coord, float dither) {
    const uint steps = 6;
    const float baseRadius = sqrt2;

    float radius    = baseRadius * (0.75 + abs(1.0-dot(sceneNormal, vec3(0.0, 1.0, 0.0))) * 0.5);

    bool hand       = depth < 0.56;
        depth       = depthLinear(depth);

    float currStep  = 0.2 * dither + 0.2;
	float fovScale  = gbufferProjection[1][1] / 1.37;
	float distScale = max((far - near) * depth + near, 5.0);
	vec2 scale      = radius * vec2(1.0 / aspectRatio, 1.0) * fovScale / distScale;

    float ao = 0.0;

    const float maxOcclusionDist    = tau;
    const float anibleedExp         = 0.71;

    for (uint i = 0; i < steps; ++i) {
		vec2 offset = offsetDist(currStep) * scale;
		float mult  = (0.7 / radius) * (far - near) * (hand ? 1024.0 : 1.0);

		float sampleDepth = depthLinear(texture(depthtex, coord + offset).r);
		float sample0 = (depth - sampleDepth) * mult;
        float antiBleed = 1.0 - rcp(1.0 + max0(distance(sampleDepth, depth) * far - maxOcclusionDist) * anibleedExp);
		float angle = mix(clamp(0.5 - sample0, 0.0, 1.0), 0.5, antiBleed);
		float dist  = mix(clamp(0.25 * sample0 - 1.0, 0.0, 1.0), 0.5, antiBleed);

		sampleDepth = depthLinear(texture(depthtex, coord - offset).r);
		sample0     = (depth - sampleDepth) * mult;
        antiBleed   = 1.0 - rcp(1.0 + max0(distance(sampleDepth, depth) * far - maxOcclusionDist) * anibleedExp);
        angle      += mix(clamp(0.5 - sample0, 0.0, 1.0), 0.5, antiBleed);
        dist       += mix(clamp(0.25 * sample0 - 1.0, 0.0, 1.0), 0.5, antiBleed);
		
		ao         += (clamp(angle + dist, 0.0, 1.0));
		currStep   += 0.2;
    }
	ao *= 1.0 / float(steps);
	
	return ao * sqrt(ao);
}

void main() {
    indirectLight   = vec4(0.0);

    vec2 lowresCoord    = uv / indirectResScale;
    ivec2 pixelPos      = ivec2(floor(uv * viewSize) / indirectResScale);
    float sceneDepth    = texelFetch(depthtex0, pixelPos, 0).x;

    GBuffer             = vec4(0, 0, 1, 1);

    if (landMask(sceneDepth) && saturate(lowresCoord) == lowresCoord) {
        vec2 uv      = saturate(lowresCoord);

        vec3 viewPos    = screenToViewSpace(vec3(uv, sceneDepth));
        vec3 viewDir    = -normalize(viewPos);

        vec4 tex1       = texelFetch(colortex1, pixelPos, 0);
        vec3 sceneNormal = decodeNormal(tex1.xy);
        vec3 viewNormal = mat3(gbufferModelView) * sceneNormal;
        vec2 lightmaps  = saturate(unpack2x8(tex1.z));

        GBuffer.xyz     = sceneNormal * 0.5 + 0.5;
        GBuffer.a       = sqrt(depthLinear(sceneDepth));
        /*
        #ifndef NO_DIRECTLIGHT
            #ifndef DIM
            vec3 directCol      = (sunAngle<0.5 ? lightColor[0] : lightColor[1]) * lightFlip;
            #elif DIM == 1
            vec3 directCol      = lightColor[0];
            #endif
        #endif
        */

        vec4 tex2       = texelFetch(colortex2, pixelPos, 0);
        int matID       = unpack2x8I(tex2.y).x;

        #ifdef ssptEnabled
            bool hand   = sceneDepth < texelFetch(depthtex2, pixelPos, 0).x;

            vec3 flatNormal     = texelFetch(colortex3, pixelPos, 0).xyz * 2.0 - 1.0;
            /*
            mat3 ssptResult     = raytraceIndirectLight(viewPos, sceneNormal, flatNormal, vec2(ditherBluenoise2(), ditherHashNoise()), lightmaps.y, hand);

            float lightmapFade  = saturate(1.0 - mix(maxOf(ssptResult[2]), avgOf(ssptResult[2]), 0.41));
            lightmapFade     = 1.0 - saturate(0.9 * (float(matID == 5) + float(matID == 6)));

            #ifndef NO_DIRECTLIGHT
            indirectLight.rgb   = ssptResult[0] + ssptResult[1] * directCol + ssptResult[2];
            #else
            indirectLight.rgb   = ssptResult[0] + ssptResult[2];
            #endif
            */

            indirectLight.rgb   = PTLight(viewPos, viewToSceneSpace(viewPos), sceneNormal, flatNormal, vec2(ditherBluenoise2(), ditherHashNoise()));

            //indirectLight.a     = lightmapFade * pow5(lightmaps.x) * ssptLightmapBlend;
        #else

            #ifdef fallbackSSAO
                float ao        = getDSSAO(depthtex0, sceneNormal, sceneDepth, uv, ditherBluenoise());
            #else
                float ao        = 1.0;
            #endif

            float groundOcclusion = exp(-max0(sceneNormal.y + 1.0) * 0.71) + 0.31;

            #ifndef DIM
            vec3 fakeBounceColor = directCol;
                fakeBounceColor *= 0.05;

            vec3 indirectColor  = fakeBounceColor * pow4(lightmaps.y) * groundOcclusion;
                indirectColor  += lightColor[1] * cube(lightmaps.y);
                indirectColor  += minimumAmbientColor * minimumAmbientMult * minimumAmbientIllum;
            #else

                #if DIM == -1
                vec3 indirectColor = lightColor[0];
                #else
                vec3 indirectColor = lightColor[1];
                #endif

            #endif

            //float lmap          = pow5(lightmaps.x);

            indirectLight.rgb   = indirectColor * ao;
            indirectLight.rgb  += ao * getBlocklightMap(blocklightColor, lightmaps.x);
            indirectLight.a     = ao;
        #endif

        //indirectLight.rgb   = debugVoxel(viewPos, flatNormal);

        flatNormals     = texelFetch(colortex3, pixelPos, 0).xyz;


    }

    indirectLight   = clamp16F(indirectLight);
}