#version 450 compatibility

/* RENDERTARGETS: 0 */
layout(location = 0) out vec3 framebuffer;

#include "/lib/head.glsl"
#include "/lib/util/encoders.glsl"

in vec2 uv;

uniform sampler2D colortex0;

uniform sampler2D depthtex0;

uniform sampler2D noisetex;

uniform sampler2D shadowtex0;
uniform sampler2D shadowtex1;

uniform sampler2D shadowcolor0;
uniform sampler2D shadowcolor1;

uniform vec2 pixelSize, viewSize;
uniform vec2 taaOffset;

uniform mat4 gbufferProjection, gbufferProjectionInverse;
uniform mat4 gbufferModelView, gbufferModelViewInverse;

uniform vec3 cameraPosition;


#define gFragment


#include "/lib/util/transforms.glsl"

#include "/lib/voxel/store.glsl"

vec3[8] octreeColor = vec3[8] (
    vec3(1.0, 0.0, 0.0)*0+1,
    vec3(1.0, 1.0, 0.0)*0+1,
    vec3(0.0, 1.0, 0.0)*0+1,
    vec3(0.0, 0.0, 1.0)*0+1,
    vec3(0.0, 0.0, 1.0),
    vec3(0.0, 1.0, 0.5),
    vec3(0.0, 1.0, 1.0),
    vec3(0.0, 0.5, 1.0)
);

vec4 octreeVis = vec4(0);
vec4 octreeVis2 = vec4(0);


ivec3 scaleIndex(ivec3 index, int targetLod) {
    vec3 x  = vec3(index) / float(targetLod);

    return ivec3(floor(x));
}


bool isVoxelOccupiedVis(ivec3 index, int level) {
    octreeVis += vec4(octreeColor[level], 1.0);

    vec4 bvh    = unpack4x4(texelFetch(shadowcolor0, getVoxelPixel(index, level), 0).a);

    if(bvh[level] > 0.5) octreeVis2 += vec4(octreeColor[level], 1.0);

    return bvh[level] > 0.5;
}

bool voxelRaytrace2(vec3 pos, ivec3 index, vec3 dir, inout vec3 octreeVis, inout int steps, int level) {
    int size    = int(pow(2, level));

    if (outsideVoxelVolume(index)) return false;

    if (isVoxelOccupiedVis(index, level)) return true;

    index   = ivec3(floor(vec3(index) / float(size)));
    pos    /= size;
    //index *= size;
    //size = 1;

    ivec3 startIndex = index;

	vec3 nextT;
	vec3 deltaDist = 1.0 / abs(dir);
	ivec3 deltaSign;

	for (int axis = 0; axis < 3; ++axis) {
		if (dir[axis] < 0.0) {
			deltaSign[axis] = -1;
			nextT[axis] = (pos[axis] - index[axis]) * deltaDist[axis];
		} else if (dir[axis] > 0.0) {
			deltaSign[axis] = 1;
			nextT[axis] = (index[axis] - pos[axis]) * deltaDist[axis] + deltaDist[axis];
		} else {
			deltaSign[axis] = 0;
			nextT[axis] = uintBitsToFloat(0x7f800000u);// = infinity
		}
	}

	if (deltaSign.x == 0 && deltaSign.y == 0 && deltaSign.z == 0) { return false; }

	bool hit = false;
	do {
		float minComp = minOf(nextT);
		if (nextT.x == minComp) {
			index.x    += deltaSign.x;
			nextT.x    += deltaDist.x;
		}
		if (nextT.y == minComp) {
			index.y    += deltaSign.y;
			nextT.y    += deltaDist.y;
		}
		if (nextT.z == minComp) {
			index.z    += deltaSign.z;
			nextT.z    += deltaDist.z;
		}

		if (outsideVoxelVolume(index * size)) break;

        hit     = getVoxelOccupancy(index * size, level);

        if (hit) octreeVis += octreeColor[level];

        steps++;
	} while (!hit);

    return hit;
}
bool marchBVH2(vec3 pos, ivec3 index, vec3 dir) {
    if (outsideVoxelVolume(index)) return false;

    ivec3 deltaSign     = ivec3(sign(dir));

    if(all(equal(deltaSign, ivec3(0)))) return false;

    ivec3 positiveDir   = max(deltaSign, ivec3(0));
    ivec3 negativeDir   = min(deltaSign, ivec3(0));

    vec3 deltaT         = 1.0 / abs(dir);

    const int maxLevel  = OCTREE_LEVELS - 1;

    vec3 steps          = (index - pos) * deltaSign + positiveDir - 1;

    int level           = 0;

    while(level++ < maxLevel && !isVoxelOccupiedVis(index, level));

    --level;

    bool hit            = false;
    int iterations      = 0;

    do {
        float minT;
        bvec3 isMinComp;

        bool potentialHit   = false;
        int octreeIterations = 0;

        if (outsideVoxelVolume(index)) break;

        do {
            int size        = int(pow(2, level));
            ivec3 plane     = (scaleIndex(index, size) + positiveDir) * size;
            vec3 planeT     = abs(plane - pos) * deltaT;

            minT            = minOf(planeT);

            if (minT > 256.0) break;

            isMinComp       = bvec3(planeT.x == minT, planeT.y == minT, planeT.z == minT);

            ivec3 prevIndex = index;

            ivec3 stepsToNext = ivec3(minT * abs(dir) - steps + 0.5 * vec3(isMinComp));
                steps      += stepsToNext;
                index      += stepsToNext * deltaSign;

            if (outsideVoxelVolume(index)) break;

            int levelAbove  = level + 1;
            int sizeAbove   = int(pow(2, levelAbove));

            bool stepUp     = scaleIndex(prevIndex, sizeAbove) != scaleIndex(index, sizeAbove);
                level       = min(stepUp ? levelAbove : level, maxLevel);

            bool blocked    = isVoxelOccupiedVis(index, level);

            while (blocked && level > 0) {
                blocked     = isVoxelOccupiedVis(index, --level);
            }

            potentialHit    = blocked && level == 0;
        } while(++octreeIterations < 64 && !potentialHit);

        if (potentialHit) hit = true;

    } while (++iterations < 64 && !hit);

    return hit;
}


vec3 voxelDebug(float sceneDepth) {
    vec3 viewPos    = screenToViewSpace(vec3(uv, sceneDepth), true);
    vec3 scenePos   = viewToSceneSpace(viewPos);

    vec3 cameraPos  = sceneToVoxelSpace(gbufferModelViewInverse[3].xyz);

    vec3 voxelPos   = sceneToVoxelSpace(scenePos);
    vec3 voxelIndex = floor(cameraPos);
    vec3 startIndex = floor(cameraPos);

    vec3 direction  = normalize(voxelPos - cameraPos);

    vec3 vis   = vec3(0.0);
    int steps   = 1;

    //bool hit    = voxelRaytrace2(cameraPos, ivec3(voxelIndex), direction, vis, steps, 0);
    //    hit    = voxelRaytrace2(cameraPos, ivec3(voxelIndex), direction, vis, steps, 1);
    //    hit    = voxelRaytrace2(cameraPos, ivec3(voxelIndex), direction, vis, steps, 2);
    //    hit    = voxelRaytrace2(cameraPos, ivec3(voxelIndex), direction, vis, steps, 3);
    bool hit    = marchBVH2(cameraPos, ivec3(voxelIndex), direction);

    return octreeVis2.rgb + vis;
}

void main() {
    framebuffer     = stex(colortex0).rgb;

        float sceneDepth    = stex(depthtex0).x;
        vec3 viewPos    = screenToViewSpace(vec3(uv, sceneDepth));
        vec3 scenePos   = viewToSceneSpace(viewPos);

        //framebuffer     = sqrt(length(viewPos)) / 16.0 + voxelDebug(sceneDepth);
}